from io import BytesIO

from welkin.models.base import Collection, Resource
from welkin.pagination import PageableIterator
from welkin.util import model_id


class DocumentSummaryFile(Resource):
    @model_id("Patient", "DocumentSummary")
    def get(self, patient_id: str, document_summary_id: str) -> BytesIO:
        response = self._client.get(
            f"{self._client.instance}/patients/{patient_id}/document-summary/"
            f"{document_summary_id}/files/{self.id}"
        )

        return BytesIO(response)


class DocumentSummaryFiles(Collection):
    resource = DocumentSummaryFile
    iterator = PageableIterator

    @model_id("Patient", "DocumentSummary")
    def create(
        self,
        patient_id: str,
        document_summary_id: str,
        files: list = None,
        *args,
        **kwargs,
    ):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/document-summary/"
            f"{document_summary_id}/files",
            files=files,
            *args,
            **kwargs,
        )


class DocumentSummary(Resource):
    subresources = [DocumentSummaryFile, DocumentSummaryFiles]

    @model_id("Patient")
    def get(self, patient_id: str):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/document-summary/{self.id}"
        )

    @model_id("Patient")
    def create(self, patient_id: str):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/document-summary"
        )

    @model_id("Patient")
    def delete(self, patient_id: str):
        return super().delete(
            f"{self._client.instance}/patients/{patient_id}/document-summary/{self.id}"
        )


class DocumentSummaries(Collection):
    resource = DocumentSummary
    iterator = PageableIterator

    @model_id("Patient")
    def get(self, patient_id: str, *args, **kwargs):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/document-summary",
            *args,
            **kwargs,
        )



from typing import Union

from welkin.models.base import Collection, Resource
from welkin.pagination import FormationIterator
from welkin.util import Target, _build_resources


class FormationBase:
    endpoint: str = None

    def __new__(cls, *args, **kwargs):
        if not cls.endpoint:
            raise AttributeError(
                f"The `endpoint` attribute must be set on {cls.__name__}"
            )

        return super().__new__(cls)


class FormationResource(FormationBase, Resource):
    def get(self, *args, **kwargs):
        return super().get(
            f"{self._formation._base_path}/{self.endpoint}/{self.name}",
            *args,
            **kwargs,
        )


class FormationCollection(FormationBase, Collection):
    resource = FormationResource
    iterator = FormationIterator

    def get(self, *args, **kwargs):
        return super().get(
            f"{self._formation._base_path}/{self.endpoint}",
            *args,
            **kwargs,
        )


class Assessment(FormationResource):
    endpoint = "assessments"


class Assessments(FormationCollection):
    resource = Assessment
    endpoint = "assessments"


class CDT(FormationResource):
    endpoint = "cdts"


class CDTs(FormationCollection):
    resource = CDT
    endpoint = "cdts"


class DocumentType(FormationResource):
    endpoint = "document-types"


class DocumentTypes(FormationCollection):
    resource = DocumentType
    endpoint = "document-types"


class EncounterDisposition(FormationResource):
    endpoint = "encounter-disposition"

    def __init__(self):
        super().__init__()

        self.name = ""


class Encounter(FormationResource):
    endpoint = "encounters"


class Encounters(FormationCollection):
    resource = Encounter
    endpoint = "encounters"


class Goal(FormationResource):
    endpoint = "goal-templates"


class Goals(FormationCollection):
    resource = Goal
    endpoint = "goal-templates"


class Program(FormationResource):
    endpoint = "programs"


class Programs(FormationCollection):
    resource = Program
    endpoint = "programs"


class Formation(Target):
    Assessment = Assessment
    Assessments = Assessments
    CDT = CDT
    CDTs = CDTs
    DocumentType = DocumentType
    DocumentTypes = DocumentTypes
    Encounter = Encounter
    EncounterDisposition = EncounterDisposition
    Encounters = Encounters
    Goal = Goal
    Goals = Goals
    Program = Program
    Programs = Programs

    def __init__(self, version: Union[int, str] = "current"):
        super().__init__()

        self._base_path = f"{self._client.instance}/formations/{version}"

        _build_resources(self, "_formation")



"""Client

This module provides a Client object to interface with the Welkin Health API.
"""
import logging
from http import HTTPStatus
from json import JSONDecodeError

from requests import HTTPError, Session
from requests.adapters import HTTPAdapter
from requests.compat import urljoin
from requests.packages.urllib3.util.retry import Retry  # type: ignore

from welkin import __version__, models
from welkin.authentication import WelkinAuth
from welkin.exceptions import WelkinHTTPError
from welkin.util import _build_resources, clean_request_params, clean_request_payload

logger = logging.getLogger(__name__)


class Client(Session):
    """Welkin Health Client.

    Constructs a :obj:`requests.Session` for Welkin Health API requests with
    authorization, base URL, request timeouts, and request retries.

    Args:
        tenant (str): Name of the organization.
        instance (str): Name of the database inside a customer space.
        api_client (str): API client name.
        secret_key (str): API client secret key.
        timeout (int, optional): :obj:`TimeoutHTTPAdapter` timeout value. Defaults to 5.
        total (int, optional): :obj:`Retry` total value. Defaults to 5.
        backoff_factor (float, optional): :obj:`Retry` backoff_factor value.
            Defaults to 0.5.

    Usage::

        from welkin import Client

        welkin = Client(
            tenant="gh",
            instance="sb-demo",
            api_client="VBOPNRYRWJIP",
            secret_key="+}B{KGTG6#zG%P;tQm0C",
        )


        ### Patient methods
        patient = welkin.Patient(firstName="Foo", lastName="Bar").create()  # Create

        patient = welkin.Patient(id="6801d498-26f4-4aee-961b-5daffcf193c8").get()  # Read
        patients = welkin.Patients().get()  # Read all/list

        patient.update(firstName="Baz")  # Update
        patient.delete()  # Delete

        ### User methods
        user = client.User(username="bar", email="bar@foo.com").create()  # Create

        user = welkin.User(id="301b2895-cbf0-4cac-b4cf-1d082faee95c").get()  # Read
        users = welkin.Users().get()  # Read all/list
        uasers = welkin.Users().get(
            search="lightmatter", region="east-coast", seat_assigned=True, user_state="ACTIVE"
        )  # Filtered read all/list

        user.update(firstName="Baz")  # Update
        user.delete()  # Delete
    """

    Assessment = models.Assessment
    AssessmentRecord = models.AssessmentRecord
    AssessmentRecordAnswers = models.AssessmentRecordAnswers
    AssessmentRecords = models.AssessmentRecords
    Assessments = models.Assessments
    CalendarEvent = models.CalendarEvent
    CalendarEvents = models.CalendarEvents
    CarePlan = models.CarePlan
    CarePlanOverview = models.CarePlanOverview
    CDT = models.CDT
    CDTs = models.CDTs
    Chat = models.Chat
    Chats = models.Chats
    DocumentSummaries = models.DocumentSummaries
    DocumentSummary = models.DocumentSummary
    DocumentSummaryFile = models.DocumentSummaryFile
    DocumentSummaryFiles = models.DocumentSummaryFiles
    Email = models.Email
    Emails = models.Emails
    Encounter = models.Encounter
    EncounterDisposition = models.EncounterDisposition
    Encounters = models.Encounters
    Formation = models.Formation
    Patient = models.Patient
    Patients = models.Patients
    PatientProgram = models.PatientProgram
    ProgramPhase = models.ProgramPhase
    ProgramPhases = models.ProgramPhases
    PatientPrograms = models.PatientPrograms
    Schedules = models.Schedules
    SearchChats = models.SearchChats
    SMS = models.SMS
    SMSes = models.SMSes
    User = models.User
    Users = models.Users
    WorkHours = models.WorkHours

    def __init__(
        self,
        tenant,
        instance,
        api_client,
        secret_key,
        timeout=5,
        total=5,
        backoff_factor=0.5,
    ):
        """Welkin Health client constructor.

        Constructs a :obj:`requests.Session` for Welkin Health API requests with
        authorization, base URL, request timeouts, and request retries.

        Args:
            tenant (str): Name of the customer space.
            instance (str): Name of the database inside a customer space.
            api_client (str): API client name.
            secret_key (str): API client secret key.
            timeout (int, optional): :obj:`TimeoutHTTPAdapter` timeout value. Defaults to 5.
            total (int, optional): :obj:`Retry` total value. Defaults to 5.
            backoff_factor (float, optional): :obj:`Retry` backoff_factor value.
                Defaults to 0.5.
        """
        super().__init__()

        self.host = f"https://api.live.welkincloud.io/{tenant}/"
        self.headers.update({"User-Agent": f"python-welkin/{__version__}"})

        adapter = TimeoutHTTPAdapter(
            timeout=timeout,
            max_retries=Retry(
                total=total,
                status_forcelist=[
                    HTTPStatus.TOO_MANY_REQUESTS,
                    HTTPStatus.INTERNAL_SERVER_ERROR,
                    HTTPStatus.BAD_GATEWAY,
                    HTTPStatus.SERVICE_UNAVAILABLE,
                    HTTPStatus.GATEWAY_TIMEOUT,
                ],
                backoff_factor=backoff_factor,
            ),
        )
        self.mount("https://", adapter)
        self.mount("http://", adapter)

        self.auth = WelkinAuth(
            tenant=tenant,
            api_client=api_client,
            secret_key=secret_key,
            token_method=self.get_token,
        )

        self.instance = instance
        _build_resources(self, "_client")

    def prepare_request(self, request):
        if request.json:
            request.json = clean_request_payload(request.json)
        if request.params:
            request.params = clean_request_params(request.params)

        return super().prepare_request(request)

    def request(
        self,
        method: str,
        path: str,
        meta_key: str = None,
        meta_dict: dict = {},
        *args,
        **kwargs,
    ):
        """Override :obj:`Session` request method to add retries and output JSON.

        Args:
            method (str): Method for the new Request object.
            path (str): Path from host for the new Request object.

        Returns:
            dict: Response JSON
        """
        if not isinstance(path, str):
            path = "/".join((str(s) for s in path if s))
        path = path.rstrip("/")

        for _ in range(2):
            response = super().request(
                method=method, url=urljoin(self.host, path), *args, **kwargs
            )

            try:
                response.raise_for_status()
                break
            except HTTPError as exc:
                code = exc.response.status_code

                if code in [HTTPStatus.UNAUTHORIZED]:
                    msg = response.json()
                    codes = ["NOT_VALID_JSON_WEB_TOKEN", "TOKEN_EXPIRED"]
                    if any(i.get("errorCode") in codes for i in msg):
                        self.auth.refresh_token()
                        continue

                raise WelkinHTTPError(exc) from exc

        try:
            json = response.json()
        except JSONDecodeError:
            if not response.content:
                return {}

            if isinstance(response.content, bytes):
                return response.content

            raise

        if "formations" in path:
            return json

        if isinstance(json, list):
            json = {
                "content": json,
                meta_key: meta_dict,
            }
        elif "rows" in json:
            json = {
                "content": json.pop("rows"),
                meta_key: meta_dict,
            }

        # Pull out the resource
        resource = json
        if "content" in json:
            resource = json.pop("content", None)
        elif "data" in json:
            resource = json.pop("data", None)

        meta = None

        # specifically with cdts the resource and metadata are both in the data dict
        if isinstance(resource, dict) and "content" in resource:
            new_resource = resource.pop("content", None)
            meta = resource.pop(meta_key, {})
            meta.update(json)
            meta.update(resource)
            resource = new_resource
        # encounter disposition formation comes as just a dictionary
        elif isinstance(resource, dict) and "formations" in path:
            resource = [resource]

        # Response metadata for pagination
        if meta_key:
            if not meta:
                meta = json.pop(meta_key, {})
            meta.update(json)
            return resource, meta

        return resource or json

    def get_token(self) -> dict:
        data = {"secret": self.auth.secret_key}
        response = self.post(f"admin/api_clients/{self.auth.api_client}", json=data)

        return response


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout, *args, **kwargs):
        """TimeoutHTTPAdapter constructor.

        Args:
            timeout (int): How many seconds to wait for the server to send data before
                giving up.
        """
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        """Override :obj:`HTTPAdapter` send method to add a default timeout."""
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout

        return super().send(request, **kwargs)



from welkin.models.assessment import (
    Assessment,
    AssessmentRecord,
    AssessmentRecordAnswers,
    AssessmentRecords,
    Assessments,
)
from welkin.models.calendar import CalendarEvent, CalendarEvents, Schedules, WorkHours
from welkin.models.care_plan import CarePlan, CarePlanOverview
from welkin.models.cdt import CDT, CDTs
from welkin.models.chat import Chat, Chats, SearchChats
from welkin.models.document import (
    DocumentSummaries,
    DocumentSummary,
    DocumentSummaryFile,
    DocumentSummaryFiles,
)
from welkin.models.email import Email, Emails
from welkin.models.encounter import Encounter, EncounterDisposition, Encounters
from welkin.models.formation import Formation
from welkin.models.patient import Patient, Patients
from welkin.models.program import (
    PatientProgram,
    PatientPrograms,
    ProgramPhase,
    ProgramPhases,
)
from welkin.models.sms import SMS, SMSes
from welkin.models.user import User, Users

__all__ = [
    "Assessment",
    "AssessmentRecord",
    "AssessmentRecordAnswers",
    "AssessmentRecords",
    "Assessments",
    "CalendarEvent",
    "CalendarEvents",
    "CarePlan",
    "CarePlanOverview",
    "CDT",
    "CDTs",
    "Chat",
    "Chats",
    "DocumentSummaries",
    "DocumentSummary",
    "DocumentSummaryFile",
    "DocumentSummaryFiles",
    "Email",
    "Emails",
    "Encounter",
    "EncounterDisposition",
    "Encounters",
    "Formation",
    "Patient",
    "Patients",
    "PatientProgram",
    "ProgramPhase",
    "ProgramPhases",
    "PatientPrograms",
    "Schedules",
    "SearchChats",
    "SMS",
    "SMSes",
    "User",
    "Users",
    "WorkHours",
]



from welkin.models.assessment import AssessmentRecord, AssessmentRecords
from welkin.models.base import Collection, Resource
from welkin.models.care_plan import CarePlan
from welkin.models.cdt import CDT, CDTs
from welkin.models.chat import Chat, Chats, SearchChats
from welkin.models.document import (
    DocumentSummaries,
    DocumentSummary,
    DocumentSummaryFile,
)
from welkin.models.email import Email, Emails
from welkin.models.encounter import Encounter, Encounters
from welkin.models.program import PatientProgram, PatientPrograms
from welkin.models.sms import SMS, SMSes
from welkin.pagination import PageableIterator


class Patient(Resource):
    subresources = [
        AssessmentRecord,
        AssessmentRecords,
        CarePlan,
        CDT,
        CDTs,
        Chat,
        Chats,
        DocumentSummaries,
        DocumentSummary,
        DocumentSummaryFile,
        Email,
        Emails,
        Encounter,
        Encounters,
        PatientProgram,
        PatientPrograms,
        SearchChats,
        SMS,
        SMSes,
    ]

    def create(self):
        return super().post(f"{self._client.instance}/patients")

    def get(self, expand: bool = None):
        _id = None
        if hasattr(self, "id"):
            type = None
            _id = self.id
        elif hasattr(self, "externalId"):
            type = "EID"
            _id = self.externalId
        elif hasattr(self, "externalGuid"):
            type = "EGUID"
            _id = self.externalGuid
        elif hasattr(self, "mrn"):
            type = "MRN"
            _id = self.mrn

        return super().get(
            f"{self._client.instance}/patients/{_id}",
            params={
                "type": type,
                expand: expand,
            },
        )

    def update(self, **kwargs):
        return super().patch(f"{self._client.instance}/patients/{self.id}", kwargs)

    def delete(self):
        return super().delete(f"{self._client.instance}/patients/{self.id}")

    def __str__(self):
        return f"{self.firstName} {self.lastName}"


class Patients(Collection):
    resource = Patient
    iterator = PageableIterator

    def get(self, filter={}, *args, **kwargs):
        # TODO: Add sort and query arguments.
        return super().post(
            f"{self._client.instance}/by-filter/patients", json=filter, *args, **kwargs
        )



from welkin.models.base import Collection, Resource
from welkin.pagination import MetaIterator
from welkin.util import model_id


class Chat(Resource):
    @model_id("Patient")
    def create(self, patient_id: str):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/chat/inbound"
        )

    def __str__(self):
        return f"{self.sender['clientType']} {self.message}"


class Chats(Collection):
    resource = Chat
    iterator = MetaIterator

    @model_id("Patient")
    def get(self, patient_id: str, include_archived: bool = False, *args, **kwargs):
        params = {
            "includeArchived": include_archived,
        }

        return super().get(
            f"{self._client.instance}/patients/{patient_id}/chat",
            params=params,
            *args,
            **kwargs,
        )


class ChatSearchResult(Resource):
    pass


class SearchChats(Collection):
    resource = ChatSearchResult
    iterator = MetaIterator

    @model_id("Patient")
    def get(
        self,
        patient_id: str,
        query: str,
        content_page_size: int = 20,
        include_archived: bool = False,
        *args,
        **kwargs,
    ):
        params = {
            "query": query,
            "contentPageSize": content_page_size,
            "includeArchived": include_archived,
        }

        return super().get(
            f"{self._client.instance}/patients/{patient_id}/chat/search",
            params=params,
            *args,
            **kwargs,
        )



from datetime import datetime

from welkin.models.base import Collection, Resource
from welkin.pagination import PageNumberIterator
from welkin.util import model_id


class CDT(Resource):
    @model_id("Patient")
    def create(self, patient_id: str):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/cdts/{self.cdtName}"
        )

    @model_id("Patient")
    def get(self, patient_id: str):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/cdts/{self.cdtName}/{self.id}"
        )

    @model_id("Patient")
    def update(self, patient_id: str, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/cdts/{self.cdtName}/{self.id}",
            kwargs,
        )

    @model_id("Patient")
    def delete(self, patient_id: str):
        return super().delete(
            f"{self._client.instance}/patients/{patient_id}/cdts/{self.cdtName}/{self.id}"
        )


class CDTs(Collection):
    resource = CDT
    iterator = PageNumberIterator

    @model_id("Patient")
    def get(
        self,
        patient_id: str,
        cdt_name: str,
        fields: list = None,
        filters: dict = None,
        date_start: datetime = None,
        date_end: datetime = None,
        sort: str = None,
        *args,
        **kwargs,
    ):
        params = {
            "fields": fields,
            "filters": filters,
            "sort": sort,
            "dateStart": date_start,
            "dateEnd": date_end,
        }

        return super().get(
            f"{self._client.instance}/patients/{patient_id}/cdts/{cdt_name}",
            params=params,
            *args,
            **kwargs,
        )

    @model_id("Patient")
    def update(self, patient_id: str, cdt_name: str, body: dict = None, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/cdts/{cdt_name}",
            json=body,
            **kwargs,
        )



from datetime import datetime, timezone
from enum import Enum

from welkin.models.base import Collection, Resource
from welkin.pagination import PageableIterator


class EventType(Enum):
    GROUP_THERAPY = "GROUP_THERAPY"
    APPOINTMENT = "APPOINTMENT"
    LEAVE = "LEAVE"
    ENCOUNTER = "ENCOUNTER"


class EventStatus(Enum):
    SCHEDULED = "SCHEDULED"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    MISSED = "MISSED"


class EventMode(Enum):
    IN_PERSON = "IN-PERSON"
    CALL = "CALL"
    VIDEO = "VIDEO"


class CalendarEvent(Resource):
    def create(self):
        # TODO: Accept User and Patient instances as participants
        return super().post(f"{self._client.instance}/calendar/events")

    def get(self):
        return super().get(f"{self._client.instance}/calendar/events/{self.id}")

    def update(self, **kwargs):
        return super().patch(
            f"{self._client.instance}/calendar/events/{self.id}", kwargs
        )

    def delete(self):
        return super().delete(f"{self._client.instance}/calendar/events/{self.id}")


class CalendarEvents(Collection):
    resource = CalendarEvent
    iterator = PageableIterator

    def get(
        self,
        from_date: datetime,
        to_date: datetime,
        participant_ids: list = None,
        event_type: str = None,
        sort: str = None,
        include_cancelled: bool = None,
        include_encounter_info: bool = None,
        exclude_assigned_to_encounter_events: bool = None,
        viewer_timezone: str = None,
        *args,
        **kwargs,
    ):
        # Validation
        if event_type:
            EventType(event_type)

        # IDEA: Consider inferring params from instance properties
        params = {
            "from": from_date,
            "to": to_date,
            "participantIds": participant_ids,
            "eventType": event_type,
            "sort": sort,
            "includeCancelled": include_cancelled,
            "includeEncounterInfo": include_encounter_info,
            "excludeAssignedToEncounterEvents": exclude_assigned_to_encounter_events,
            "viewerTimezone": viewer_timezone,
        }

        return super().get(
            f"{self._client.instance}/calendar/events", params=params, *args, **kwargs
        )


class Schedule(Resource):
    pass


class Schedules(Collection):
    resource = Schedule
    iterator = PageableIterator

    def get(
        self,
        ids: list,
        from_date: datetime,
        to_date: datetime,
        include_cancelled: bool = None,
        available: bool = False,
        full: bool = False,
        *args,
        **kwargs,
    ):
        route = "psm-schedules"
        if available:
            route = "available-psm-schedules"
        if full:
            route = "full-psm-schedules"

        params = {
            "psmIds": ids,
            "from": from_date,
            "to": to_date,
            "includeCancelled": include_cancelled,
        }

        return super().get(
            f"{self._client.instance}/calendar/{route}", params=params, *args, **kwargs
        )


class WorkHours(Collection):
    iterator = PageableIterator

    def get(
        self,
        from_date: datetime,
        to_date: datetime,
        psm_ids: list = None,
        timezone: str = None,
        *args,
        **kwargs,
    ):
        params = {
            "from": from_date,
            "to": to_date,
            "psm-ids": psm_ids,
            "viewerTimezone": timezone,
        }

        response = self._client.get(
            f"{self._client.instance}/calendar/work-hours/",
            *args,
            params=params,
            **kwargs,
        )

        # When no work hours are found Welkin returns an {None: []} dictionary
        if isinstance(response, dict) and None in response.keys():
            return WorkHours([])

        return response



from datetime import date, datetime, timezone
from functools import lru_cache, wraps
from typing import Any, Callable, Tuple, Union
from uuid import UUID

import inflection

from welkin.models.base import Collection, Resource, SchemaBase

# NOTE: `clean_request_payload` and `clean_request_params` are intentionally DRY
# violations. The code may be the same, but they represent different knowledge.
# the same goes for `clean_request_payload` and `clean_json_list`


class Target:
    _client = None

    def __init__(self):
        self._base_path = ""
        _build_resources(self, "_client", self._client)


def _build_resources(instance: type, attribute_name: str, value: type = None) -> None:
    """Add an attribute pointing to an instance for each resource.

    Args:
        instance (type): _description_
        attribute_name (str): _description_
        value (type, optional): _description_. Defaults to None.
    """
    if not value:
        value = instance

    for attribute in dir(instance):
        val = getattr(instance, attribute)

        try:
            if not issubclass(val, (SchemaBase, Target)):
                continue
        except TypeError:
            continue  # Failed because `issubclass` expects a class.

        setattr(val, attribute_name, value)


def clean_data(value: Any) -> Any:
    """Clean data for JSON serialization.

    Args:
        value (Any): The value to clean.

    Returns:
        Any: The cleaned value.
    """
    if isinstance(value, datetime):
        return clean_datetime(value)
    elif isinstance(value, date):
        return clean_date(value)
    elif isinstance(value, dict):
        return clean_request_payload(value)
    elif isinstance(value, list):
        return clean_json_list(value)
    elif isinstance(value, UUID):
        return str(value)

    # No cleaning needed
    return value


def clean_request_payload(payload: dict) -> dict:
    result = {}
    for k, v in payload.items():
        result[k] = clean_data(v)

    return result


def clean_json_list(data: list) -> list:
    result = []
    for item in data:
        result.append(clean_data(item))

    return result


def clean_request_params(params: dict) -> dict:
    result = {}
    for k, v in params.items():
        if isinstance(v, datetime):
            result[k] = clean_datetime(v)
        elif isinstance(v, date):
            result[k] = clean_date(v)
        elif isinstance(v, list):
            result[k] = ",".join(v)
        else:
            result[k] = v

    return result


def clean_date(date: date) -> str:
    dt = datetime(date.year, date.month, date.day, 0, 0, 0, tzinfo=timezone.utc)

    return clean_datetime(dt)


def clean_datetime(dt: datetime) -> str:
    return (
        dt.astimezone(tz=timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )


def find_model_id(instance: Union[Collection, Resource], model_name: str) -> str:
    """Recursively traverse the `_parent` chain searching for a model id.

    Args:
        instance (Union[Collection, Resource]): The instanceect instance to inspect.
        model_name (str): The class name of the model to find.

    Raises:
        AttributeError: If recursion ends without finding the model id.

    Returns:
        str: The model id.
    """
    body_id_key = f"{to_camel_case(model_name)}Id"

    if instance.__class__.__name__ == model_name:
        return instance.id
    elif hasattr(instance, body_id_key):
        return getattr(instance, body_id_key)
    elif instance._parent is not None:
        return find_model_id(instance._parent, model_name)

    raise AttributeError(
        f"Cannot find {model_name} id. Model._parent chain ends in {instance}"
    )


def model_id(*models: Tuple[str]) -> Callable:
    """Insert values for `model_id` arguments if not provided.

    Args:
        *models (Tuple[str]): The model names to search for.

    Raises:
        TypeError: If no ID is found and no arguments are provided.
    """

    def decorator(f: Callable):
        @wraps(f)
        def wrapper(self, *args, **kwargs) -> Callable:
            outer_exc = None
            for model in models:
                key = f"{to_snake_case(model)}_id"

                if not args and key not in kwargs:
                    try:
                        kwargs[key] = find_model_id(self, model)
                    except AttributeError as e:
                        try:
                            raise e from outer_exc
                        except AttributeError as exc:
                            outer_exc = exc

            try:
                return f(self, *args, **kwargs)
            except TypeError as exc:
                # Raise from the outer `AttributeError` so we don't lose context.
                raise exc from outer_exc

        return wrapper

    return decorator


@lru_cache(maxsize=None)
def to_camel_case(s: str) -> str:
    """Convert a string to camelCase.

    Args:
        s (str): The string to convert.

    Returns:
        str: The converted camelCase string.
    """
    return inflection.camelize(to_snake_case(s), uppercase_first_letter=False)


@lru_cache(maxsize=None)
def to_snake_case(s: str) -> str:
    """Convert a string to snake_case.

    Args:
        s (str): The string to convert.

    Returns:
        str: The converted snake_case string.
    """
    return inflection.underscore(s)



from welkin.models.base import Collection, Resource
from welkin.pagination import PageableIterator
from welkin.util import model_id


class SMS(Resource):
    @model_id("Patient")
    def get(self, patient_id: str, *args, **kwargs):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/sms/{self.id}",
            *args,
            **kwargs,
        )


class SMSes(Collection):
    resource = SMS
    iterator = PageableIterator

    @model_id("Patient")
    def get(self, patient_id: str, sort: str = None, *args, **kwargs):
        params = {
            "sort": sort,
        }

        return super().get(
            f"{self._client.instance}/patients/{patient_id}/sms",
            params=params,
            *args,
            **kwargs,
        )



import sys

from welkin.pagination import PageIterator


class SchemaBase:
    _client = None
    _parent = None
    subresources = []

    def __getattr__(self, name):
        try:
            for r in self.subresources:
                if r.__name__ == name:
                    r._parent = self
                    return r

        except AttributeError:
            pass

        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            raise AttributeError(e) from None


class Resource(dict, SchemaBase):
    nested_objects = {}

    def __getattr__(self, name):
        try:
            value = super().__getitem__(name)
        except KeyError:
            value = super().__getattr__(name)

        if name in self.nested_objects:
            cls = getattr(sys.modules["welkin.models"], self.nested_objects[name])
            if issubclass(cls, Collection):
                value = cls(cls.resource(item) for item in value)
            else:
                value = cls(value)

        return value

    def __setattr__(self, name, value):
        super().__setitem__(name, value)

    def __delattr__(self, name):
        super().__delitem__(name)

    def __str__(self):
        id = getattr(self, "id", "")

        return f"{self.__class__.__name__} #{id}" if id else self.__class__.__name__

    def __repr__(self):
        return object.__repr__(self)

    def get(self, resource, subresource=None, *args, **kwargs):
        response = self._client.get([resource, subresource], *args, **kwargs)
        super().update(response)

        return self

    def patch(self, resource, data, *args, **kwargs):
        response = self._client.patch(
            resource,
            json=data,
            *args,
            **kwargs,
        )

        super().update(response)

        return self

    def post(self, resource, *args, **kwargs):
        response = self._client.post(
            resource,
            json=self,
            *args,
            **kwargs,
        )
        super().update(response)

        return self

    def put(self, resource, *args, **kwargs):
        response = self._client.put(
            resource,
            json=self,
            *args,
            **kwargs,
        )

        super().update(response)

        return self

    def delete(self, resource, *args, **kwargs):
        response = self._client.delete(resource, *args, **kwargs)
        super().update(response)

        return self


class Collection(list, SchemaBase):
    resource = Resource
    iterator = PageIterator

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__class__(*super().__getitem__(index))

        return super().__getitem__(index)

    def __str__(self):
        return str([str(i) for i in self])

    def __repr__(self):
        return object.__repr__(self)

    def get(self, *args, **kwargs):
        return self.request(self._client.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request(self._client.post, *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.request(self._client.patch, *args, **kwargs)

    def request(self, method, resource, paginate=False, *args, **kwargs):
        paginator = self.iterator(self, resource, method, *args, **kwargs)

        if paginate:
            return paginator

        self.clear()
        for n, _ in enumerate(paginator):
            if n == paginator.size - 1:
                break

        return self



from welkin.models.base import Resource
from welkin.util import model_id


class CarePlanOverview(Resource):
    @model_id("Patient")
    def update(self, patient_id: str):
        return super().put(
            f"{self._client.instance}/patients/{patient_id}/care-plan/overview"
        )


class CarePlan(Resource):
    subresources = [CarePlanOverview]

    @model_id("Patient")
    def create(self, patient_id: str):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/care-plan/overview"
        )

    @model_id("Patient")
    def get(self, patient_id: str):
        return super().get(f"{self._client.instance}/patients/{patient_id}/care-plan")



from enum import Enum

from welkin.models.base import Collection, Resource
from welkin.models.encounter import Encounters
from welkin.pagination import MetaInfoIterator


class User(Resource):
    subresources = [Encounters]

    def __str__(self):
        try:
            return f"{self.firstName} {self.lastName}"
        except AttributeError:
            return self.username

    def create(self):
        return super().post("admin/users")

    def get(self):
        return super().get(f"admin/users/{self.id}", params=dict(type="ID"))

    def update(self, **kwargs):
        return super().patch(f"admin/users/{self.username}", kwargs)

    def delete(self):
        return super().delete(f"admin/users/{self.id}", params=dict(type="ID"))


class UserState(Enum):
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    PROVISIONED = "PROVISIONED"


class Users(Collection):
    resource = User
    iterator = MetaInfoIterator

    def get(
        self,
        search: str = None,
        region: str = None,
        seat_assigned: bool = None,
        user_state: str = None,
        *args,
        **kwargs,
    ):
        # TODO: Figure out sort arguments
        params = dict(
            search=search,
            seatAssigned=seat_assigned,
            userState=user_state,
        )

        # User state validation
        if user_state:
            UserState(user_state)

        path = "admin/users"
        if region:
            path = f"{self._client.instance}/users"
            params["region"] = region

        return super().get(path, params=params, *args, **kwargs)



"""Package version information."""
from datetime import date
from importlib.metadata import distribution

_dist_metadata: dict = distribution("welkin").metadata.json

__title__ = _dist_metadata["name"]
__description__ = _dist_metadata["summary"]
__url__ = _dist_metadata["home_page"]
__version__ = _dist_metadata["version"]
__author__ = _dist_metadata["author"]
__author_email__ = _dist_metadata["author_email"]
__license__ = _dist_metadata["license"]
__copyright__ = f"{date.today().year} Lightmatter"



from welkin.models.base import Collection, Resource
from welkin.pagination import PageableIterator
from welkin.util import model_id


class Assessment(Resource):
    @model_id("Patient", "Encounter")
    def create(self, patient_id: str, encounter_id: str):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}"
            "/assessments"
        )

    @model_id("Patient", "Encounter")
    def get(self, patient_id: str, encounter_id: str):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            f"assessments/{self.id}"
        )

    @model_id("Patient", "Encounter")
    def update(self, patient_id: str, encounter_id: str, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            f"assessments/{self.id}",
            kwargs,
        )

    @model_id("Patient", "Encounter")
    def delete(self, patient_id: str, encounter_id: str):
        return super().delete(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            f"assessments/{self.id}"
        )


class Assessments(Collection):
    resource = Assessment
    iterator = PageableIterator

    @model_id("Patient", "Encounter")
    def get(self, patient_id: str, encounter_id: str, *args, **kwargs):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/"
            "assessments",
            *args,
            **kwargs,
        )


class AssessmentRecordAnswers(Resource):
    @model_id("Patient", "AssessmentRecord")
    def update(self, patient_id: str, assessment_record_id: str):
        return super().put(
            f"{self._client.instance}/patients/{patient_id}/"
            f"assessment-records/{assessment_record_id}/answers"
        )


class AssessmentRecord(Resource):
    subresources = [AssessmentRecordAnswers]

    @model_id("Patient")
    def create(self, patient_id: str):
        return super().post(
            f"{self._client.instance}/patients/{patient_id}/assessment-records"
        )

    @model_id("Patient")
    def get(self, patient_id: str):
        return super().get(
            f"{self._client.instance}/patients/"
            f"{patient_id}/assessment-records/{self.id}"
        )

    @model_id("Patient")
    def update(self, patient_id: str):
        return super().put(
            f"{self._client.instance}/patients/"
            f"{patient_id}/assessment-records/{self.id}"
        )

    @model_id("Patient")
    def delete(self, patient_id: str):
        return super().delete(
            f"{self._client.instance}/patients/"
            f"{patient_id}/assessment-records/{self.id}",
        )


class AssessmentRecords(Collection):
    resource = AssessmentRecord
    iterator = PageableIterator

    @model_id("Patient")
    def get(self, patient_id: str, **kwargs):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/assessment-records",
            **kwargs,
        )



from welkin.models.base import Collection, Resource
from welkin.pagination import MetaInfoIterator
from welkin.util import model_id


class ProgramPhase(Resource):
    pass


class ProgramPhases(Collection):
    resource = ProgramPhase


class PatientProgram(Resource):
    subresources = [ProgramPhase]
    nested_objects = {
        "phases": "ProgramPhase",
        "currentPhase": "ProgramPhase",
        "pathHistory": "ProgramPhases",
    }

    @model_id("Patient")
    def get(
        self,
        patient_id: str,
        assigned_programs: bool = None,
        sort: str = None,
        *args,
        **kwargs,
    ):
        path = f"{self._client.instance}/patients/{patient_id}/programs"
        if hasattr(self, "id"):
            path += f"/history/{self.id}"
        elif hasattr(self, "programName"):
            path += f"/current/{self.programName}"
        else:
            raise ValueError("Program must have either programName or id attributes")

        return super().get(
            path,
            params={
                "assignedPrograms": assigned_programs,
                "sort": sort,
            },
            *args,
            **kwargs,
        )

    @model_id("Patient")
    def update(self, patient_id: str, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/programs/"
            f"{self.programName}",
            kwargs,
        )

    @model_id("Patient")
    def delete(self, patient_id: str):
        return super().delete(
            f"{self._client.instance}/patients/{patient_id}/programs/{self.id}"
        )


class PatientPrograms(Collection):
    resource = PatientProgram
    iterator = MetaInfoIterator

    @model_id("Patient")
    def get(
        self,
        patient_id: str,
        assigned_programs: bool = None,
        sort: str = None,
        *args,
        **kwargs,
    ):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/programs",
            params={
                "assignedPrograms": assigned_programs,
                "sort": sort,
            },
            *args,
            **kwargs,
        )



from enum import Enum

from welkin.models.assessment import Assessment, Assessments
from welkin.models.base import Collection, Resource
from welkin.pagination import MetaInfoIterator
from welkin.util import model_id


class EncounterDisposition(Resource):
    @model_id("Patient", "Encounter")
    def get(self, patient_id: str, encounter_id: str):
        return super().get(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/disposition"
        )

    @model_id("Patient", "Encounter")
    def update(self, patient_id: str, encounter_id: str, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/encounters/{encounter_id}/disposition",
            kwargs,
        )


class EncounterStatus(Enum):
    OPEN = "OPEN"
    CANCELLED = "CANCELLED"
    DRAFT = "DRAFT"
    FINALIZED = "FINALIZED"


class Encounter(Resource):
    subresources = [Assessment, Assessments, EncounterDisposition]
    nested_objects = {
        "assessmentLinks": "Assessments",
        "userRelatedToCalendarEvent": "User",
        "disposition": "EncounterDisposition",
    }

    @model_id("Patient")
    def create(self, patient_id: str):
        return super().post(f"{self._client.instance}/patients/{patient_id}/encounters")

    @model_id("Patient")
    def get(self, patient_id: str, related_data: bool = False):
        encounters = "encounters"
        if related_data:
            encounters = "full-encounters"

        return super().get(
            f"{self._client.instance}/patients/{patient_id}/{encounters}/{self.id}"
        )

    @model_id("Patient")
    def update(self, patient_id: str, **kwargs):
        return super().patch(
            f"{self._client.instance}/patients/{patient_id}/encounters/{self.id}",
            kwargs,
        )

    @model_id("Patient")
    def delete(self, patient_id: str):
        return super().delete(
            f"{self._client.instance}/patients/{patient_id}/encounters/{self.id}"
        )


class Encounters(Collection):
    resource = Encounter
    iterator = MetaInfoIterator

    def get(
        self,
        patient_id: str = None,
        user_id: str = None,
        related_data: bool = None,
        with_care_team: bool = None,
        only_with_calendar_event: bool = None,
        statuses: list = None,
        sort: str = None,
        *args,
        **kwargs,
    ):
        path = f"{self._client.instance}/"
        if patient_id:
            path += f"patients/{patient_id}/"
        elif user_id:
            path += f"users/{user_id}/"
        elif self._parent.__class__.__name__ == "Patient":
            path += f"patients/{self._parent.id}/"
        elif self._parent.__class__.__name__ == "User":
            path += f"users/{self._parent.id}/"

        path += "full-encounters" if related_data else "encounters"

        params = {
            "withCareTeam": with_care_team,
            "statuses": statuses,
            "sort": sort,
            "onlyWithCalendarEvent": only_with_calendar_event,
        }

        return super().get(path, params=params, *args, **kwargs)



import logging
import shelve
import tempfile
from pathlib import Path
from typing import Callable

from portalocker import Lock
from requests import PreparedRequest
from requests.auth import AuthBase

logger = logging.getLogger(__name__)


DB_PATH = str(Path(tempfile.gettempdir(), ".welkin.db"))
DB_LOCK = f"{DB_PATH}.lock"


class WelkinAuth(AuthBase):
    """Attaches API Key Authentication to the given Request object.

    https://developers.welkinhealth.com/#authentication
    """

    def __init__(
        self,
        tenant: str,
        api_client: str,
        secret_key: str,
        token_method: Callable[[], dict],
    ) -> None:
        self.tenant = tenant
        self.api_client = api_client
        self.secret_key = secret_key
        self.token_method = token_method

    def __eq__(self, other: object) -> bool:
        return (self.tenant, self.api_client, self.secret_key) == (
            getattr(other, "tenant", None),
            getattr(other, "api_client", None),
            getattr(other, "secret_key", None),
        )

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        logger.info(f"{r.method} {r.url}")

        if "api_clients" not in r.path_url:
            r.headers["Authorization"] = f"Bearer {self.token}"

        return r

    @property
    def token(self) -> str:
        with Lock(DB_LOCK):
            with shelve.open(DB_PATH) as db:
                try:
                    return db[self.tenant]["token"]
                except KeyError:
                    pass

        self.refresh_token()

        return self.token

    @token.setter
    def token(self, value: dict) -> None:
        with Lock(DB_LOCK):
            with shelve.open(DB_PATH) as db:
                db[self.tenant] = value

    def refresh_token(self) -> None:
        self.token = self.token_method()
